import asyncio
from collections.abc import Callable, Awaitable
from typing import Never

from .auth import auth
from .connection import Connection
from .transport.stream import StreamTransport


async def serve(
    host: str,
    port: int,
    privkey_b: bytes,
    peer_pubkey_b: bytes,
    callback: Callable[[Connection], Awaitable[None | Never]]
) -> None:

    async def handle(
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter
    ) -> None:
        try:
            transport = StreamTransport(reader, writer)
            connection = await auth(transport, privkey_b, peer_pubkey_b)
            await callback(connection)
        finally:
            writer.close()
            await writer.wait_closed()

    server = await asyncio.start_server(handle, host, port)
    async with server:
        await server.serve_forever()



from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey
)
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey
)
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from .connection import Connection
from .protocol.message import MessageProtocol
from .protocol.signed import SignedProtocol
from .transport.abc import Transport


async def auth(
    transport: Transport,
    privkey_b: bytes,
    peer_pubkey_b: bytes
) -> Connection:
    privkey = Ed25519PrivateKey.from_private_bytes(privkey_b)
    peer_pubkey = Ed25519PublicKey.from_public_bytes(peer_pubkey_b)
    return await _auth(transport, privkey, peer_pubkey)


async def _auth(
    transport: Transport,
    privkey: Ed25519PrivateKey,
    peer_pubkey: Ed25519PublicKey
) -> Connection:
    auth_key = await _exchange_auth_key(transport, privkey, peer_pubkey)
    protocol = MessageProtocol(auth_key)
    connection = Connection(transport, protocol)
    return connection


async def _exchange_auth_key(
    transport: Transport,
    privkey: Ed25519PrivateKey,
    peer_pubkey: Ed25519PublicKey
) -> bytes:
    protocol = SignedProtocol(privkey, peer_pubkey)
    connection = Connection(transport, protocol)

    dh_privkey = X25519PrivateKey.generate()
    dh_pubkey = dh_privkey.public_key()

    dh_pubkey_b = dh_pubkey.public_bytes_raw()
    await connection.send(dh_pubkey_b)

    peer_dh_pubkey_b = await connection.receive()
    peer_dh_pubkey = X25519PublicKey.from_public_bytes(peer_dh_pubkey_b)

    shared_key = dh_privkey.exchange(peer_dh_pubkey)

    kdf = HKDF(hashes.SHA3_256(), 32, None, None)
    auth_key = kdf.derive(shared_key)

    return auth_key



import os
import struct

from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305

from .abc import Protocol


class MessageProtocol(Protocol):
    """
    Scheme:
    -------------------------------------------
    | Nonce | Authentication Tag + Ciphertext |
    -------------------------------------------
    Plaintext (decrypted Ciphertext) scheme:
    -----------------
    | Seq | Payload |
    -----------------
    """
    _SEQ_STRUCT = struct.Struct('<Q')

    def __init__(self,
        auth_key: bytes
    ) -> None:
        self._aead = ChaCha20Poly1305(auth_key)
        self._send_seq = 0
        self._recv_seq = 0

    def pack(self,
        payload: bytes
    ) -> bytes:
        seq = self._get_send_seq()
        seq_b = self._SEQ_STRUCT.pack(seq)
        plaintext = seq_b + payload

        nonce = os.urandom(12)
        ciphertext = self._aead.encrypt(nonce, plaintext, None)
        data = nonce + ciphertext

        return data

    def unpack(self,
        data: bytes
    ) -> bytes:
        nonce, ciphertext = data[:12], data[12:]
        plaintext = self._aead.decrypt(nonce, ciphertext, None)

        seq_b, payload = plaintext[:self._SEQ_STRUCT.size], \
            plaintext[self._SEQ_STRUCT.size:]
        seq, = self._SEQ_STRUCT.unpack(seq_b)
        assert self._get_recv_seq() == seq

        return payload

    def _get_send_seq(self) -> int:
        self._send_seq += 1
        return self._send_seq

    def _get_recv_seq(self) -> int:
        self._recv_seq += 1
        return self._recv_seq



from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey


def generate_private_key() -> bytes:
    privkey = Ed25519PrivateKey.generate()
    privkey_b = privkey.private_bytes_raw()
    return privkey_b


def get_public_key(
    privkey_b: bytes,
    /
) -> bytes:
    privkey = Ed25519PrivateKey.from_private_bytes(privkey_b)
    pubkey = privkey.public_key()
    pubkey_b = pubkey.public_bytes_raw()
    return pubkey_b



import asyncio
import struct

from .abc import Transport


class StreamTransport(Transport):
    """
    --------------------
    | Length | Payload |
    --------------------
    """
    _LENGTH_STRUCT = struct.Struct('<Q')

    def __init__(self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter
    ) -> None:
        self._reader = reader
        self._writer = writer

    async def send(self,
        payload: bytes
    ) -> None:
        length_b = self._LENGTH_STRUCT.pack(len(payload))
        data = length_b + payload
        self._writer.write(data)
        await self._writer.drain()

    async def receive(self) -> bytes:
        length_b = await self._reader.readexactly(self._LENGTH_STRUCT.size)
        length, = self._LENGTH_STRUCT.unpack(length_b)
        payload = await self._reader.readexactly(length)
        return payload



from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey
)

from .abc import Protocol


class SignedProtocol(Protocol):
    """
    Scheme:
    -----------------------
    | Signature | Payload |
    -----------------------
    """

    def __init__(self,
        privkey: Ed25519PrivateKey,
        peer_pubkey: Ed25519PublicKey
    ) -> None:
        self._privkey = privkey
        self._peer_pubkey = peer_pubkey

    def pack(self,
        payload: bytes
    ) -> bytes:
        signature = self._privkey.sign(payload)
        data = signature + payload
        return data

    def unpack(self,
        data: bytes
    ) -> bytes:
        signature, payload = data[:64], data[64:]
        self._peer_pubkey.verify(signature, payload)
        return payload



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from .auth import auth
from .connection import Connection
from .transport.stream import StreamTransport


@asynccontextmanager
async def connect(
    host: str,
    port: int,
    privkey_b: bytes,
    peer_pubkey_b: bytes
) -> AsyncIterator[Connection]:
    reader, writer = await asyncio.open_connection(host, port)
    try:
        transport = StreamTransport(reader, writer)
        connection = await auth(transport, privkey_b, peer_pubkey_b)
        yield connection
    finally:
        writer.close()
        await writer.wait_closed()



from .protocol.abc import Protocol
from .transport.abc import Transport


class Connection:

    def __init__(self,
        transport: Transport,
        protocol: Protocol
    ) -> None:
        self._transport = transport
        self._protocol = protocol

    async def send(self,
        payload: bytes
    ) -> None:
        data = self._protocol.pack(payload)
        await self._transport.send(data)

    async def receive(self) -> bytes:
        data = await self._transport.receive()
        payload = self._protocol.unpack(data)
        return payload



import abc


class Transport(abc.ABC):

    @abc.abstractmethod
    async def send(self,
        payload: bytes
    ) -> None:
        ...

    @abc.abstractmethod
    async def receive(self) -> bytes:
        ...



import abc


class Protocol(abc.ABC):

    @abc.abstractmethod
    def pack(self,
        payload: bytes
    ) -> bytes:
        ...

    @abc.abstractmethod
    def unpack(self,
        data: bytes
    ) -> bytes:
        ...
